@@ -1,6 +1,7 @@
 /* System-dependent calls for tar.
 
-   Copyright 2003-2008, 2010, 2013-2014 Free Software Foundation, Inc.
+   Copyright 2003-2008, 2010, 2013-2014, 2016 Free Software Foundation,
+   Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -330,6 +331,7 @@ sys_child_open_for_compress (void)
   pid_t grandchild_pid;
   pid_t child_pid;
 
+  signal (SIGPIPE, SIG_IGN);
   xpipe (parent_pipe);
   child_pid = xfork ();
 
